import collections.abc
import typing
import pydantic
from .interaction import Message
from .state import StateScope, StateIdentifierType


class ContextModel(pydantic.BaseModel):
    pass


ContextDataType = ContextModel | collections.abc.Mapping[StateIdentifierType, ContextModel]


class ContextCollector(typing.Protocol):
    async def __call__(
        self,
        model_contract: type[ContextModel],
        state_scope: StateScope,
        message: Message | None = None
    ) -> collections.abc.Mapping[StateIdentifierType, ContextModel]: ...



import collections.abc
import typing

from .identity import IdentifierType
from .interaction import Message, Command, MessageStreamType
from .operations import Actor
from .state import StateModel

MessageActorMap = collections.abc.MutableMapping[
    type[Message],
    set[tuple[type[Actor], type[StateModel] | None]]
]


class CommandProcessor(typing.Protocol):
    async def __aenter__(self) -> 'CommandProcessor': ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    async def receive(self, command: Command, scope: collections.abc.Sequence[IdentifierType] | IdentifierType): ...

    def get_message_stream(self) -> MessageStreamType: ...

    async def commit(self): ...

    async def shutdown(self): ...


class CommandProcessorSettings(typing.Protocol):
    register_events: bool = True
    auto_commit: bool = True


class CommandProcessorFactory(typing.Protocol):
    def get_processor(
        self,
        settings: CommandProcessorSettings
    ) -> CommandProcessor: ...



import abc
import collections.abc
import inspect

from ..definitions import contracts
from ..exploration import search_for_subtype


class StateKeeper(contracts.StateKeeper, abc.ABC):
    @classmethod
    def __advertising__(cls) -> collections.abc.Iterable[type[contracts.StateModel]]:
        for member in inspect.getmembers(cls, predicate=inspect.isfunction):
            for _param_type in member[1].__annotations__.values():
                if _state_model_contract := search_for_subtype(_param_type, contracts.StateModel):
                    yield _state_model_contract



import typing
import dataclasses
import collections.abc
from zorge import contracts as zorge_contracts
from .operations import Actor
from .interaction import Message
from .state import StateModel, StateHandler
from .context import ContextModel, ContextCollector

DependencyContainer = zorge_contracts.DependencyContainer


@dataclasses.dataclass
class DependencyMapValue:
    field_name: str
    field_type: type[typing.Any]


@dataclasses.dataclass
class DependencyMaps:
    message_actor: dict[
        type[Message],
        type[Actor]
    ]
    actor_state: dict[
        type[Actor],
        DependencyMapValue
    ]
    actor_context: dict[
        type[Actor],
        DependencyMapValue
    ]
    message_context: dict[
        type[Message],
        type[ContextModel]
    ]
    state_getter: dict[
        type[StateModel],
        DependencyMapValue
    ]
    state_keeper: dict[
        type[StateModel],
        DependencyMapValue
    ]
    context_model_collector: dict[
        type[ContextModel],
        DependencyMapValue
    ]


class DependencyRegistry:
    def register_connection_context(
        self,
        open_callback: typing.Callable,
        close_success_callback: typing.Callable,
        close_failure_callback: typing.Callable,
        contract: zorge_contracts.DependencyBindingContract
    ): ...

    def register_registry(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract
    ): ...

    def register_data_vendor(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract
    ): ...

    def register_context_collector(
        self,
        instance: type[ContextCollector],
        contract: type[ContextModel]
    ): ...

    def register_actor(
        self,
        instance: type[Actor],
    ): ...

    def register_state_handler(
        self,
        instance: type[StateHandler]
    ): ...

    def register_view(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract | None = None,
        scope: typing.Literal['instance', 'global'] = 'instance'
    ): ...

    def get_container(self) -> zorge_contracts.DependencyContainer: ...

    def get_maps(self) -> DependencyMaps: ...


class StateRegistry(typing.Protocol):
    def register_state(
        self,
        state_model_contract: type[StateModel],
        state_handler_contract: type[StateHandler]
    ): ...

    def get_handler(
        self,
        state_model_contract: type[StateModel]
    ) -> type[StateHandler] | None: ...


class ContextRegistry(typing.Protocol):
    def register_context(
        self,
        context_model_contract: type[ContextModel],
        context_collector_contract: type[ContextCollector]
    ): ...

    def get_collector(
        self,
        context_model_contract: type[ContextModel]
    ) -> type[ContextCollector] | None: ...


class MessageRegistry(typing.Protocol):
    def register_message(
        self,
        message_contract: type[Message],
        actor_contract: type[Actor],
        context_model_contract: type[ContextModel] | None = None
    ): ...

    def get_actors_with_context(
        self,
        message_contract: type[Message]
    ) -> collections.abc.Iterable[tuple[type[Actor], type[ContextModel] | None]]: ...



import typing


def search_for_subtype(given_type: type[typing.Any], desired_type: type[typing.Any]):
    if not given_type or not desired_type:
        return
    if hasattr(given_type, '__mro__') and desired_type in given_type.__mro__:
        return given_type
    for _type in typing.get_args(given_type):
        if result := search_for_subtype(_type, desired_type):
            return result



import collections.abc


class PanchoException(Exception):
    def __init__(
        self,
        contract: type = None,
        message: str = None,
        context: collections.abc.Mapping | None = None
    ):
        self.contract = contract
        self.message = message
        self.context = context
        super().__init__()


class MarkupException(PanchoException):
    pass


class ActorException(PanchoException):
    pass


class ActorInitializationException(ActorException):
    pass


class ActorMessageProcessingException(ActorException):
    pass


class DependencyException(PanchoException):
    pass


class ContextException(PanchoException):
    pass


class LoadContextException(ContextException):
    pass


class StateException(PanchoException):
    pass


class CommitStateException(StateException):
    pass



from .definitions import (
    ActorException,
    MarkupException,
    StateException,
    ContextException,
    CommitStateException,
    DependencyException,
    LoadContextException,
    ActorInitializationException,
    ActorMessageProcessingException
)



import abc
import collections.abc
import inspect
import typing

from ..definitions import contracts
from ..exploration.types import search_for_subtype


class Actor(contracts.Actor, abc.ABC):
    __internals: contracts.ActorInternals = None

    def __init__(
        self,
        state_id: contracts.StateIdentifierType | None = None
    ):
        self._state_id = state_id
        self._state = None

    def __receive__(
        self,
        message: contracts.Message,
        context: contracts.ContextModel | None = None
    ) -> collections.abc.Iterable[contracts.Event] | None:
        handler_method_name = self._search_for_instance(type(message))
        if handler_method_name:
            method = getattr(self, handler_method_name)
            if self.__internals__().message_context_map.get(type(message)):
                events = method(message, context)
            else:
                events = method(message)
            if events is not None:
                if not isinstance(events, collections.abc.Sequence):
                    events = (events,)
                return events

    @abc.abstractmethod
    def __state__(self) -> contracts.StateModel | None:
        ...

    def __state_id__(self) -> contracts.IdentifierType | None:
        return self._state_id

    @classmethod
    def __internals__(cls) -> contracts.ActorInternals:
        if not cls.__internals:
            message_handler_map = {}
            message_context_map = {}
            for member in inspect.getmembers(cls, predicate=inspect.isfunction):
                _method = member[1].__dict__.get('__wrapped__') or member[1]
                _message_contract = None
                _annotations = _method.__annotations__
                if _annotations:
                    for v in _annotations.values():
                        if hasattr(v, '__mro__') and (contracts.Event in v.__mro__ or contracts.Command in v.__mro__):
                            _message_contract = v
                            message_handler_map[_message_contract] = member[0]
                            break

                    if _message_contract is None:
                        continue

                    for v in _annotations.values():
                        if context_model_contract := search_for_subtype(v, contracts.ContextModel):
                            message_context_map[_message_contract] = context_model_contract

            cls.__internals = contracts.ActorInternals(
                message_handler_map=message_handler_map,
                message_context_map=message_context_map
            )
        return cls.__internals

    @classmethod
    def __advertising__(cls) -> typing.Iterable[
        tuple[type[contracts.Message] | type[contracts.ContextModel]]
    ]:
        for message_contract in cls.__internals__().message_handler_map.keys():
            context_contract = cls.__internals__().message_context_map.get(message_contract)
            yield message_contract, context_contract

    def _search_for_instance(self, message_contract) -> str | None:
        for contract in message_contract.__mro__:
            if instance := self.__internals__().message_handler_map.get(contract):
                return instance



import dataclasses
import typing
import collections.abc

from pancho.definitions import contracts


class MessageQueue:
    def __init__(
        self,
        frames: collections.abc.Iterable[contracts.MessageFrame] | None = None
    ):
        self._data = frames or []
        self._index = 0

    def enqueue(
        self,
        data: contracts.MessageFrame
    ) -> typing.NoReturn:
        self._data.append(data)

    def dequeue(self) -> contracts.MessageFrame | None:
        if self._index >= len(self._data):
            return
        result = self._data[self._index]
        self._index += 1
        return result

    def get_stream(self) -> contracts.MessageStreamType:
        return (m.message for m in self._data)


@dataclasses.dataclass
class ProcessorSettings:
    register_events: bool = True
    auto_commit: bool = True



import functools
from .instance import Processor
from .generic import ProcessorSettings
from ..definitions import contracts
from ..bootstraping.registering import MessageRegistry, StateRegistry, ContextRegistry


class ProcessorFactory(contracts.CommandProcessorFactory):
    def __init__(
        self,
        dependency_registry: contracts.DependencyRegistry,
    ):
        self._dependency_container = dependency_registry.get_container()

    def get_processor(
        self,
        settings: contracts.CommandProcessorSettings = ProcessorSettings()
    ) -> Processor:
        return Processor(
            dependency_container=self._dependency_container,
            message_registry=self._message_registry,
            state_registry=self._state_registry,
            context_registry=self._context_registry,
            settings=settings
        )

    @functools.cached_property
    def _message_registry(self) -> contracts.MessageRegistry:
        dc = self._dependency_container
        mr = MessageRegistry()
        for actor_contract in dc.get_bindings().filter(lambda c, r: contracts.Actor in c.__mro__).keys():
            for message_contract, context_model_contract in actor_contract.__advertising__():
                mr.register_message(
                    message_contract=message_contract,
                    actor_contract=actor_contract,
                    context_model_contract=context_model_contract
                )
        return mr

    @functools.cached_property
    def _state_registry(self) -> contracts.StateRegistry:
        dc = self._dependency_container
        sr = StateRegistry()
        for state_handler_contract in dc.get_bindings().filter(lambda c, r: contracts.StateHandler in c.__mro__).keys():
            for state_model_contract in state_handler_contract.__advertising__() or ():
                sr.register_state(
                    state_model_contract=state_model_contract,
                    state_handler_contract=state_handler_contract
                )
        return sr

    @functools.cached_property
    def _context_registry(self) -> contracts.ContextRegistry:
        dc = self._dependency_container
        cr = ContextRegistry()
        for context_collector_contract in dc.get_bindings().filter(
            lambda c, r: contracts.ContextCollector in c.__mro__
        ).keys():
            for context_model_contract in context_collector_contract.__advertising__():
                cr.register_context(
                    context_model_contract=context_model_contract,
                    context_collector_contract=context_collector_contract
                )
        return cr



import collections.abc
import typing

from zorge.di.container import DependencyContainer
from zorge import contracts as zorge_contracts
from ..definitions import contracts
from ..identity.factory import UUIDIdentifierFactory
from ..interaction.factory import InteractionFactory


class DependencyRegistry:
    def __init__(
        self,
        dc: zorge_contracts.DependencyContainer | None = None,
        identifier_factory: contracts.IdentifierFactory = UUIDIdentifierFactory,
        interaction_factory: contracts.InteractionFactory = InteractionFactory,
    ):
        self._dc = dc or DependencyContainer()
        self._dc.register_contractual_dependency(
            instance=identifier_factory,
            contract=contracts.IdentifierFactory
        )
        self._dc.register_contractual_dependency(
            instance=interaction_factory,
            contract=contracts.InteractionFactory
        )

    def register_connection_context(
        self,
        open_callback: typing.Callable,
        close_success_callback: typing.Callable,
        close_failure_callback: typing.Callable,
        contract: zorge_contracts.DependencyBindingContract
    ):
        self._dc.register_instance_singleton(
            instance=open_callback,
            contract=contract,
        )
        self._dc.register_shutdown_callback(
            success_callback=close_success_callback,
            failure_callback=close_failure_callback,
            contract=contract,
        )

    def register_registry(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract
    ):
        self._dc.register_global_singleton(
            instance=instance,
            contract=contract
        )

    def register_data_vendor(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract
    ):
        self._dc.register_contractual_dependency(
            instance=instance,
            contract=contract
        )

    def register_context_collector(
        self,
        instance: type[contracts.ContextCollector],
    ):
        self._dc.register_selfish_dependency(
            instance=instance,
        )

    def register_actor(
        self,
        instance: type[contracts.Actor],
    ):
        self._dc.register_selfish_dependency(
            instance=instance,
            singleton_scope=zorge_contracts.DependencyBindingScope.INSTANCE
        )

    def register_state_handler(
        self,
        instance: type[contracts.StateHandler]
    ):
        self._dc.register_selfish_dependency(
            instance=instance,
            singleton_scope=zorge_contracts.DependencyBindingScope.INSTANCE
        )

    def register_view(
        self,
        instance: zorge_contracts.DependencyBindingInstance,
        contract: zorge_contracts.DependencyBindingContract | None = None,
        scope: typing.Literal['instance', 'global'] = 'instance'
    ):
        if not contract:
            self._dc.register_selfish_dependency(
                instance=instance,
                singleton_scope={
                    'instance': zorge_contracts.DependencyBindingScope.INSTANCE,
                    'global': zorge_contracts.DependencyBindingScope.GLOBAL
                }[scope]
            )
        else:
            {
                'instance': self._dc.register_global_singleton,
                'global': self._dc.register_global_singleton
            }[scope](
                instance=instance,
                contract=contract
            )

    def get_container(self) -> DependencyContainer:
        return self._dc


class StateRegistry(contracts.StateRegistry):
    def __init__(self):
        self._map = {}

    def register_state(
        self,
        state_model_contract: type[contracts.StateModel],
        state_handler_contract: type[contracts.StateHandler]
    ):
        if contracts.StateKeeper in state_handler_contract.__mro__:
            self._map[state_model_contract] = state_handler_contract

    def get_handler(
        self,
        state_model: type[contracts.StateModel]
    ) -> type[contracts.StateHandler] | None:
        return self._map.get(state_model)


class ContextRegistry(contracts.ContextRegistry):
    def __init__(self):
        self._map = {}

    def register_context(
        self,
        context_model_contract: type[contracts.ContextModel],
        context_collector_contract: type[contracts.ContextCollector]
    ):
        self._map[context_model_contract] = context_collector_contract

    def get_collector(
        self,
        context_model: type[contracts.ContextModel]
    ) -> type[contracts.ContextCollector] | None:
        return self._map.get(context_model)


class MessageRegistry(contracts.MessageRegistry):
    def __init__(self):
        self._map = collections.defaultdict(set)

    def register_message(
        self,
        message_contract: type[contracts.Message],
        actor_contract: type[contracts.Actor],
        context_model_contract: type[contracts.ContextModel] | None = None
    ):
        self._map[message_contract].add((actor_contract, context_model_contract))

    def get_actors_with_context(
        self,
        message_contract: type[contracts.Message]
    ) -> collections.abc.Iterable[tuple[type[contracts.Actor], type[contracts.ContextModel] | None]]:
        for item in self._map.get(self._search_for_contract(message_contract), ()):
            yield item

    def _search_for_contract(self, message_contract) -> type:
        for contract in message_contract.__mro__:
            if contract in self._map:
                return contract



import collections.abc
import inspect
import typing

from ..definitions import contracts
from ..exploration.types import search_for_subtype


class ContextCollector(contracts.ContextCollector):
    __internals = {}

    async def __call__(
        self,
        model_contract: type[contracts.ContextModel],
        state_scope: contracts.StateScope,
        message: contracts.Message | None = None
    ) -> collections.abc.Mapping[contracts.StateIdentifierType, contracts.ContextModel]:
        if method := self._search_for_instance(model_contract):
            data = await method(self, state_scope=state_scope, message=message)
            if not isinstance(state_scope, collections.abc.Sequence) and not isinstance(data, collections.abc.Mapping):
                return {
                    state_scope: data
                }
            return data

    @classmethod
    def __internals__(cls):
        if not cls.__internals:
            for member in inspect.getmembers(cls, predicate=inspect.isfunction):
                _method = member[1].__dict__.get('__wrapped__') or member[1]
                _annotations = _method.__annotations__ or {}
                if context_model_contract := search_for_subtype(_annotations.get('return'), contracts.ContextModel):
                    cls.__internals[context_model_contract] = _method
        return cls.__internals

    @classmethod
    def __advertising__(cls):
        return (c for c in cls.__internals__().keys())

    def _search_for_instance(
        self,
        model_contract: type[contracts.ContextModel]
    ) -> typing.Callable:
        for contract in model_contract.__mro__:
            if instance := self.__internals__().get(contract):
                return instance



import pydantic
import collections.abc
import typing

from .identity import IdentifierType

StateIdentifierType = IdentifierType


class StateModel(pydantic.BaseModel):
    pass


T = typing.TypeVar('T', bound=StateModel)
StateDataType = T | collections.abc.Iterable[T] | None

StateScope = collections.abc.Sequence[StateIdentifierType] | StateIdentifierType


class StateHandler(typing.Protocol):
    @classmethod
    def __advertising__(cls) -> collections.abc.Iterable[type[StateModel]]: ...


class StateKeeper(StateHandler):
    async def __call__(self, state: StateDataType) -> typing.NoReturn: ...



import collections.abc
import typing
import datetime
import pydantic
import dataclasses

from .identity import IdentifierType
from .state import StateScope


class Message(pydantic.BaseModel):
    created_at: datetime.datetime = pydantic.Field(default_factory=datetime.datetime.utcnow)


class Command(Message):
    launch_at: datetime.datetime | None = pydantic.Field(default=None)


class Event(Message):
    state_id: IdentifierType


class InfoEvent(Event):
    context: dict | None = pydantic.Field(default=None)


class ExceptionContext(pydantic.BaseModel):
    code: int
    message: str
    details: list | None = pydantic.Field(default=None)


class ExceptionEvent(Event):
    context: ExceptionContext


class CommandProducer(typing.Protocol):
    def bind(
        self,
        command_contract: type[Command],
        clause: typing.Callable[[], bool] | None = None,
        fields_map: collections.abc.Mapping[str, str] | None = None
    ): ...

    def get(self) -> typing.Generator[type[Command], None, None]: ...


class EventProducer(typing.Protocol):
    def bind(
        self,
        event_contract: type[Event]
    ): ...

    def get(self) -> typing.Generator[type[Event], None, None]: ...


class InteractionFactory(typing.Protocol):
    def get_command_producer(
        self,
        actor_id: IdentifierType | None = None,
        payload: collections.abc.Mapping | None = None
    ) -> CommandProducer: ...

    def get_event_producer(
        self,
        actor_id: IdentifierType,
        context: collections.abc.Mapping | None = None
    ) -> EventProducer: ...


@dataclasses.dataclass
class MessageFrame:
    message: Message
    state_scope: StateScope | None = None


class MessageQueue(typing.Protocol):
    def enqueue(
        self,
        data: MessageFrame | collections.abc.Iterable[MessageFrame]
    ) -> typing.NoReturn:
        ...

    def dequeue(
        self
    ) -> MessageFrame | None:
        ...


MessageStreamType = typing.Iterable[Message]



from .identity import (
    IdentifierType,
    IdentifierFactory
)
from .vendoring import (
    DataVendor,
    DefaultDataVendor,
    EventRegistratorDataVendor,
    EventRegistratorFrameName
)
from .operations import (
    Actor,
    ActorInternals
)
from .interaction import (
    Message,
    Event,
    Command,
    InteractionFactory,
    MessageStreamType,
    MessageFrame,
    MessageQueue,
    ExceptionEvent,
    InfoEvent,
    ExceptionContext
)
from .processing import (
    MessageActorMap,
    CommandProcessorFactory,
    CommandProcessor,
    CommandProcessorSettings
)
from .state import (
    StateIdentifierType,
    StateDataType,
    StateScope,
    StateHandler,
    StateKeeper,
    StateModel,
)
from .context import (
    ContextDataType,
    ContextModel,
    ContextCollector
)
from .obtaining import (
    ListViewData,
    EntryViewData,
    LiteralViewData,
    View,
    ViewFactory
)
from .markup import (
    FilterSchema,
    SortSchema,
    PaginationSchema
)
from .registering import (
    DependencyRegistry,
    DependencyContainer,
    DependencyMaps,
    DependencyMapValue,
    ContextRegistry,
    StateRegistry,
    MessageRegistry
)



import collections.abc
import functools
import typing

import zorge.contracts
from zorge.contracts import DependencyContainer
from pancho.definitions import contracts, exceptions
from .generic import MessageQueue
from ..exploration import search_for_subtype


class ActorIdentityMap:
    def __init__(self):
        self._actors = collections.defaultdict(dict)

    def register(
        self,
        state_id: contracts.IdentifierType,
        instance: contracts.Actor
    ):
        self._actors[instance.__class__.__name__][state_id] = instance

    def get_one(
        self,
        contract: type[contracts.Actor],
        state_id: contracts.IdentifierType
    ) -> contracts.Actor | None:
        return self._actors.get(contract.__name__, {}).get(state_id)

    def get_all(self) -> dict[str, dict[contracts.IdentifierType, contracts.Actor]]:
        return self._actors


class Processor(contracts.CommandProcessor):
    def __init__(
        self,
        dependency_container: DependencyContainer,
        message_registry: contracts.MessageRegistry,
        context_registry: contracts.ContextRegistry,
        state_registry: contracts.StateRegistry,
        settings: contracts.CommandProcessorSettings
    ):
        self._dependency_container = dependency_container
        self._message_registry = message_registry
        self._context_registry = context_registry
        self._state_registry = state_registry
        self._settings = settings

    async def __aenter__(self) -> 'Processor':
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._settings.auto_commit and not exc_type:
            await self.commit()

        await self._dependency_provider.shutdown(exc_type)

    def get_message_stream(self) -> contracts.MessageStreamType:
        return self._message_queue.get_stream()

    async def receive(
        self,
        command: contracts.Command,
        state_scope: contracts.StateScope
    ):
        self._message_queue.enqueue(
            contracts.MessageFrame(
                message=command,
                state_scope=state_scope if isinstance(state_scope, collections.abc.Sequence) else (state_scope,)
            )
        )
        while message_frame := self._message_queue.dequeue():
            _message = message_frame.message
            _state_scope = message_frame.state_scope or (None,)
            for actor_contract, message_context_contract in self._message_registry.get_actors_with_context(
                type(_message)
            ):
                context = await self._load_context(
                    context_model_contract=message_context_contract,
                    state_scope=_state_scope,
                    message=_message
                )

                for state_id in _state_scope:
                    for event in self._process_message(
                        actor=await self._get_actor(
                            contract=actor_contract,
                            state_id=state_id
                        ),
                        context=context,
                        message=_message,
                        state_id=state_id
                    ) or ():
                        self._message_queue.enqueue(
                            contracts.MessageFrame(message=event, state_scope=(state_id,))
                        )

    async def commit(self):
        for actors in self._actor_identity_map.get_all().values():
            state = []
            for actor in actors.values():
                state.append(actor.__state__())
            await self._keep_state(state)

    async def shutdown(self, exc_type: typing.Any | None = None):
        await self._dependency_provider.shutdown(exc_type)

    async def _get_actor(
        self,
        contract: type[contracts.Actor],
        state_id: contracts.IdentifierType | None = None
    ) -> contracts.Actor:
        actor = self._actor_identity_map.get_one(contract, state_id)
        if not actor:
            params = {}
            for k, v in contract.__init__.__annotations__.items():
                if search_for_subtype(v, contracts.IdentifierType):
                    params[k] = state_id
                    break
            try:
                actor = contract(**params)
            except Exception as e:
                raise exceptions.ActorInitializationException(contract=contract) from e
            self._actor_identity_map.register(state_id, actor)
        return actor

    @staticmethod
    def _process_message(
        actor: contracts.Actor,
        message: contracts.Message,
        state_id: contracts.IdentifierType,
        context: contracts.ContextModel
    ):
        try:
            return actor.__receive__(
                message=message,
                context=context.get(state_id) if isinstance(context, collections.abc.Mapping) else context
            )
        except Exception as e:
            raise exceptions.ActorMessageProcessingException(type(actor)) from e

    async def _keep_state(
        self,
        state: contracts.StateDataType
    ) -> typing.NoReturn:
        if not state:
            return

        state_model_contract = type(state)
        if isinstance(state, collections.abc.Sequence):
            state_model_contract = type(state[0])

        keeper_contract = self._state_registry.get_handler(state_model_contract)
        if not keeper_contract:
            return

        try:
            keeper = await self._dependency_provider.resolve(keeper_contract)
        except Exception as e:
            raise exceptions.DependencyException(contract=keeper_contract) from e

        try:
            await keeper(state)
        except Exception as e:
            raise exceptions.CommitStateException(contract=keeper_contract) from e

    async def _load_context(
        self,
        context_model_contract: type[contracts.ContextModel],
        state_scope: contracts.StateScope,
        message: contracts.Message
    ) -> contracts.StateDataType | None:
        if not context_model_contract:
            return

        collector_contract = self._context_registry.get_collector(context_model_contract)
        if not collector_contract:
            return

        try:
            collector = await self._dependency_provider.resolve(collector_contract)
        except Exception as e:
            raise exceptions.DependencyException(contract=collector_contract) from e

        try:
            return await collector(context_model_contract, state_scope, message)
        except Exception as e:
            raise exceptions.LoadContextException(contract=type(message)) from e

    @functools.cached_property
    def _dependency_provider(self) -> zorge.contracts.DependencyProvider:
        return self._dependency_container.get_provider()

    @functools.cached_property
    def _message_queue(self) -> MessageQueue:
        return MessageQueue()

    @functools.cached_property
    def _actor_identity_map(self):
        return ActorIdentityMap()



import typing
import collections.abc
import dataclasses

from .interaction import Message
from .context import ContextModel
from .state import StateModel


@dataclasses.dataclass
class ActorInternals:
    message_handler_map: dict = dataclasses.field(default_factory=dict)
    message_context_map: dict = dataclasses.field(default_factory=dict)


@typing.runtime_checkable
class Actor(typing.Protocol):
    def __receive__(
        self,
        message: Message,
        context: ContextModel | None = None
    ) -> collections.abc.Iterable[Message] | None: ...

    def __state__(self) -> StateModel | None: ...

    @classmethod
    def __advertising__(cls) -> typing.Iterable[
        tuple[type[Message] | type[ContextModel]]
    ]: ...
